// Copyright (C) 2006 Michel Santos
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

/**
 * Created on August 18, 2006, 3:13 PM
 */

package netx.jnlp.cache;


import java.io.File;
import java.net.URL;
import java.util.Date;
import netx.jnlp.Version;
import netx.jnlp.runtime.JNLPRuntime;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.HeadMethod;
import org.apache.commons.httpclient.params.DefaultHttpParamsFactory;
import org.apache.commons.httpclient.params.HttpClientParams;
import org.apache.commons.httpclient.params.HttpMethodParams;
import org.apache.commons.httpclient.params.HttpParams;


/**
 * This tracker only downloads resources if they either have no cache entry or if they have changed
 *
 * @author Michel Santos
 * @since 0.51
 */
public class SmartResourceTracker extends ResourceTracker {
  /**
   * Creates a resource tracker that does not prefetch resources.
   */
  public SmartResourceTracker() {
    this(false);
  }
  
  /**
   * Creates a resource tracker.
   *
   * @param prefetch whether to download resources before requested.
   */
  public SmartResourceTracker(boolean prefetch) {
    this.prefetch = prefetch;
    
    if (prefetch) {
      synchronized (prefetchTrackers) {
        prefetchTrackers.add(this);
        prefetchTrackers.trimToSize();
      }
    }
  }
  
  
  /**
   * Open a URL connection and get the content length and other
   * fields.
   */
  protected void initializeResource(Resource resource) {
    resource.fireDownloadEvent(); // fire CONNECTING
    
    try {
      File localFile = CacheUtil.getCacheFile(resource.location, resource.downloadVersion);
      
      // connect
      if (JNLPRuntime.isDebug()) {
        System.out.println("Opening connection to: " + resource.location);
      }
      
      HttpParams params = new DefaultHttpParamsFactory().getDefaultParams();
      params.setParameter(HttpMethodParams.USER_AGENT, "JNLP/1.5 NetX/0.51");
      HttpClient client = new HttpClient(new HttpClientParams(params));
      HeadMethod head = new HeadMethod(resource.location.toString());
      long lLastModified = 0;
      long lContentLength = 0;
      try {
        client.executeMethod(head);
        
        Header headerLastModified = head.getResponseHeader("last-modified");
        if (headerLastModified != null) {
          String strLastModified = headerLastModified.getValue();
          java.text.SimpleDateFormat dfDate = new java.text.SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss z");
          Date dateLastModified = dfDate.parse(strLastModified);
          lLastModified = dateLastModified.getTime();
        }
          
        Header headerContentLength = head.getResponseHeader("content-length");
        if (headerContentLength != null) {
          String strContentLength = headerContentLength.getValue();
          lContentLength = Long.parseLong(strContentLength);
        }
        
      } finally {
        head.releaseConnection();
      }
      
      boolean current = isCurrent(resource.location, resource.requestVersion, lLastModified, lContentLength);
      
      synchronized(resource) {
        resource.localFile = localFile;
        
        if (!current) {
          if (JNLPRuntime.isDebug()) {
            System.out.println("Cache Status: not current; downloading: " + resource.location);
          }
          resource.connection = resource.location.openConnection(); // this won't change so should be okay unsynchronized          
          resource.size = resource.connection.getContentLength();
          resource.changeStatus(CONNECT|CONNECTING, CONNECTED);
          
        } else {
          if (JNLPRuntime.isDebug()) {
            System.out.println("Cache Status: current; using: " + resource.location);
          }
          // check if up-to-date; if so set as downloaded
          resource.connection = null;
          resource.size = lContentLength;
          resource.changeStatus(DOWNLOAD|DOWNLOADING, DOWNLOADED);
        }
      }
      
      // update cache entry
      CacheEntry entry = new CacheEntry(resource.location, resource.requestVersion);
      if (!current) {
//        entry.initialize(connection);
        entry.setLastModified(Long.toString(lLastModified));
        entry.setContentLength(Long.toString(lContentLength));
      }
      
      entry.setLastUpdated(System.currentTimeMillis());
      entry.store();
      
      synchronized(lock) {
        lock.notifyAll(); // wake up wait's to check for completion
      }
      resource.fireDownloadEvent(); // fire CONNECTED
    } catch (Exception ex) {
      if (JNLPRuntime.isDebug())
        ex.printStackTrace();
      
      resource.changeStatus(0, ERROR);
      synchronized(lock) {
        lock.notifyAll(); // wake up wait's to check for completion
      }
      resource.fireDownloadEvent(); // fire ERROR
    }
  }
  
  
  /**
   * CacheUtil<p>
   * Returns whether there is a version of the URL contents in the
   * cache and it is up to date.  This method may not return
   * immediately.
   *
   * @param source the source URL
   * @param version the versions to check for
   *
   * @return whether the cache contains the version
   * @throws IllegalArgumentException if the source is not cacheable
   */
  public static boolean isCurrent(URL source, Version version, long currentLastModified, long currentContentLength) {
    if (!CacheUtil.isCacheable(source, version)) {
      throw new IllegalArgumentException(CacheUtil.R("CNotCacheable", source));
    }
    
    CacheEntry entry = new CacheEntry(source, version); // could pool this
    
    boolean result = entry.isCurrent(currentLastModified);
    
    if (JNLPRuntime.isDebug()) {
      System.out.println("isCurrent: " + source + " = " + result);
    }
    
    return result;
  }
  
}